# **********************************************************************
# Copyright 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import json
import nsx_logging
from cli.utils.command_utils import invoke_command
from cli.utils.node_utils import get_node_type
from cli.utils.constants import NODE_TYPE_HOST_CENTOSKVM
from cli.utils.constants import NODE_TYPE_HOST_SLESKVM
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import NODE_TYPE_HOST_RHELKVM
from cli.utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from cli.utils.constants import NODE_TYPE_HOST_WIN
from cli.utils.constants import NODE_TYPE_HOST_PUBLIC_CLOUD
from cli.utils.constants import NODE_TYPE_BAREMETAL_UBUNTU_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_SLES_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_OEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_WINDOWS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_CONTAINER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_CONTAINER
from cli.utils.location_utils import FilePaths
from cli.utils.error_code import ErrorCode as ec
from cli.commands.cli_command_provider import CliCommandProvider

lg = nsx_logging.getLogger("cli.commands.host_shared.intelligence_flows")


#
# NSX Intelligence flows related commands
#
_PACE_FLOWS_CLI = {
    NODE_TYPE_HOST_SLESKVM: {
        'bin': "/usr/bin/ovs-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_HOST_CENTOSKVM: {
        'bin': "/usr/bin/ovs-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_HOST_ESX: {
        'bin': "/opt/vmware/nsx-cli/bin/nsx-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_HOST_RHELKVM: {
        'bin': "/usr/bin/ovs-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_HOST_UBUNTUKVM: {
        'bin': "/usr/bin/ovs-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_HOST_PUBLIC_CLOUD: {
        'bin': FilePaths.OVS_APPCTL_PATH,
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_HOST_WIN: {
        'bin': FilePaths.OVS_APPCTL_PATH,
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_BAREMETAL_UBUNTU_SERVER: {
        'bin': "/usr/bin/ovs-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_BAREMETAL_RHEL_SERVER: {
        'bin': "/usr/bin/ovs-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_BAREMETAL_CENTOS_SERVER: {
        'bin': "/usr/bin/ovs-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_BAREMETAL_SLES_SERVER: {
        'bin': "/usr/bin/ovs-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_BAREMETAL_OEL_SERVER: {
        'bin': "/usr/bin/ovs-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_BAREMETAL_RHEL_CONTAINER: {
        'bin': "/usr/bin/ovs-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_BAREMETAL_CENTOS_CONTAINER: {
        'bin': "/usr/bin/ovs-appctl",
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    NODE_TYPE_BAREMETAL_WINDOWS_SERVER: {
        'bin': FilePaths.OVS_APPCTL_PATH,
        'socket': FilePaths.EXPORTER_CLI_SOCKET
        },
    }


def has_pace_flows_cli(node_type):
    return (node_type == NODE_TYPE_HOST_SLESKVM or
            node_type == NODE_TYPE_HOST_CENTOSKVM or
            node_type == NODE_TYPE_HOST_ESX or
            node_type == NODE_TYPE_HOST_RHELKVM or
            node_type == NODE_TYPE_HOST_UBUNTUKVM or
            node_type == NODE_TYPE_BAREMETAL_UBUNTU_SERVER or
            node_type == NODE_TYPE_BAREMETAL_RHEL_SERVER or
            node_type == NODE_TYPE_BAREMETAL_CENTOS_SERVER or
            node_type == NODE_TYPE_BAREMETAL_OEL_SERVER or
            node_type == NODE_TYPE_BAREMETAL_SLES_SERVER or
            node_type == NODE_TYPE_BAREMETAL_WINDOWS_SERVER or
            node_type == NODE_TYPE_BAREMETAL_RHEL_CONTAINER or
            node_type == NODE_TYPE_BAREMETAL_CENTOS_CONTAINER or
            node_type == NODE_TYPE_HOST_WIN)


def invoke_pace_flows_command(cmd_list):
    node_type = get_node_type()
    if has_pace_flows_cli(node_type):
        cmd = [_PACE_FLOWS_CLI[node_type]['bin'], "-t", _PACE_FLOWS_CLI[node_type]['socket']]
        cmd.extend(cmd_list)
        rc, out, err = invoke_command(cmd)
        return (rc, out, err)
    else:
        lg.error("node type %s does not have NSX Intelligence flows CLI", node_type,
                 error_code=ec.INVALID_ARGUMENT)
        return (-1, None, "invalid command")


def get_pace_flows_cli_output(command):
    cmd_list = command
    rc, out, err = invoke_pace_flows_command(cmd_list)
    if rc != 0:
        lg.error("Failed to invoke NSX Intelligence flows CLI command, rc: %d, err: %s", rc, err,
                 error_code=ec.OPERATION_FAILED)
        out = None
    return out


class PaceShowFlowsBase(CliCommandProvider):
   CMD_STR = None
   CMD_ARG = None
   def __call__(self, cmd_list):
      cmd_result = ""
      cmds = [self.CMD_STR]
      if self.CMD_ARG:
         cmds.extend([self.CMD_ARG])
      res = get_pace_flows_cli_output(cmds)
      if res is None:
          cmd_result = "MSG_PACE_FLOWS_COMM_ERROR"
      elif "not available" in res:
          cmd_result = "MSG_PACE_FLOWS_NOT_AVAILABLE"
      else:
          cmd_result = "OK\n\n%s" % res

      return cmd_result


class PaceShowFlowsConfig(PaceShowFlowsBase):
    CMD_STR = "get/intelcfg"
    CMD_ARG = None


class PaceShowFlowsMask(PaceShowFlowsBase):
    CMD_STR = "get/intelmasks"
    CMD_ARG = None


class PaceShowFlowsStats(PaceShowFlowsBase):
    CMD_STR = "get/intelstats"
    CMD_ARG = None


class PaceShowFlowsStatsAck(PaceShowFlowsBase):
    CMD_STR = "get/intelstats"
    CMD_ARG = "ack"


class PaceClearFlowsStats(PaceShowFlowsBase):
    CMD_STR = "clr/intelstats"
    CMD_ARG = None
