# ***************************************************************************
# Copyright 2014-2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
try:
    import simplejson
except ImportError:
    import json as simplejson
import signal

from cli.commands.cli_command_provider import CliCommandProvider
from cli.commands.common.arguments import get_interface_names
from cli.utils.constants import CAPTURE_STARTED
from cli.utils.constants import CAPTURE_FAILED
from cli.utils.constants import CAPTURE_INVALID_INTERFACE
from cli.utils.constants import CAPTURE_INTERRUPT_CTRL_C
from cli.utils.constants import NODE_TYPE_EDGE
from cli.utils.constants import NODE_TYPE_AUTONOMOUS_EDGE
from cli.utils.constants import NODE_TYPE_PUBLIC_CLOUD_GATEWAY
from cli.utils.constants import NODE_TYPE_HOST_CENTOSKVM
from cli.utils.constants import NODE_TYPE_HOST_SLESKVM
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import NODE_TYPE_HOST_RHELKVM
from cli.utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from cli.utils.node_utils import get_node_type
from cli.utils.node_utils import is_baremetal
from cli.utils.pcap_utils import interface_locked
from cli.utils.pcap_utils import lock_interface
from cli.utils.pcap_utils import start_packet_capture
from cli.utils.pcap_utils import stop_packet_capture_to_file
from cli.utils.pcap_utils import stop_all_packet_capture_to_file
from cli.utils.pcap_utils import unlock_interface


lg = nsx_logging.getLogger("cli.commands.ops.pcap")


g_node_type = get_node_type()
if g_node_type in [NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE, NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
    from cli.commands.edge.ops.packet_capture import PacketCaptureCommandRunner
elif ((g_node_type == NODE_TYPE_HOST_SLESKVM) or
      (g_node_type == NODE_TYPE_HOST_CENTOSKVM) or
      (g_node_type == NODE_TYPE_HOST_RHELKVM) or
      (g_node_type == NODE_TYPE_HOST_UBUNTUKVM) or
      is_baremetal()):
    from cli.commands.host_kvm.ops.packet_capture \
        import PacketCaptureCommandRunner


sigint_handler = signal.getsignal(signal.SIGINT)
sigterm_handler = signal.getsignal(signal.SIGTERM)
sigpipe_handler = signal.getsignal(signal.SIGPIPE)


class SetPacketCapture(CliCommandProvider):
    def __call__(self, cmd_list):
        try:
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            signal.signal(signal.SIGPIPE, signal.SIG_IGN)
            session = self.get_arg_value_by_keyword("session", cmd_list)
            dvfilter_name = self.get_arg_value_by_keyword("dvfilter", cmd_list)
            dvfilter_stage = self.get_arg_value_by_keyword("stage", cmd_list)
            direction = self.get_arg_value_by_keyword("direction", cmd_list)
            snaplen = self.get_arg_value_by_keyword("snaplen", cmd_list)
            core = self.get_arg_value_by_keyword("core", cmd_list)
            if session:
                interface_id = "span-%s" % session
                lock_interface_id = interface_id
            else:
                interface_id = self.get_arg_value_by_keyword("interface", cmd_list)
                lock_interface_id = interface_id
                # For edge, we reserve session 0 to create span session
                if g_node_type in [
                        NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE, NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
                    local_interfaces = get_interface_names()
                    if interface_id not in local_interfaces:
                        edge_session = PacketCaptureCommandRunner()
                        # there may be a stale session from previous pcap
                        # try at-least twice to create span session before returning error
                        # rc is true on success
                        for i in range(2):
                            rc, rv = edge_session.create_default_edge_span_session(
                                        interface_id, direction, core)
                            if rc:
                                break
                            edge_session.destroy_default_edge_span_session()
                        else:
                            return rv
                        interface_id = edge_session.default_span_interface
                    elif core:
                        return "MSG_SPAN_CORE_ERROR_LOCAL_INTERFACE"
                elif ((g_node_type == NODE_TYPE_HOST_SLESKVM) or
                      (g_node_type == NODE_TYPE_HOST_CENTOSKVM) or
                      (g_node_type == NODE_TYPE_HOST_RHELKVM) or
                      (g_node_type == NODE_TYPE_HOST_UBUNTUKVM) or
                      is_baremetal()):
                    local_interfaces = get_interface_names()
                    if interface_id not in local_interfaces:
                        kvm_session = PacketCaptureCommandRunner()
                        rv = (
                            kvm_session.create_default_kvm_span_session(
                                interface_id, direction))
                        if not rv.startswith("OK"):
                            return rv
                        interface_id = kvm_session.default_span_port
                elif g_node_type == NODE_TYPE_HOST_ESX:
                  if not interface_id:
                     if dvfilter_name:
                        # we treat dvfilter's name as interface id
                        interface_id = dvfilter_name
                     else:
                        # we treat "trace" as a special inteface id
                        interface_id = 'trace'
            # Per PM, only allow one capture to a .pcap file for a given interface
            # at a time, we manage this by requesting a lock for the interface
            if not lock_interface(lock_interface_id):
                return ("MSG_OPS_PACKET_CAPTURE_ALREADY_STARTED_ERROR\n%s" %
                        lock_interface_id)
            # In KVM, open connection to the span_port
            if ((g_node_type == NODE_TYPE_HOST_SLESKVM) or
                (g_node_type == NODE_TYPE_HOST_CENTOSKVM) or
                (g_node_type == NODE_TYPE_HOST_RHELKVM) or
                (g_node_type == NODE_TYPE_HOST_UBUNTUKVM) or
                is_baremetal()):
                kvm_session = PacketCaptureCommandRunner()
                rc = kvm_session.open_connection_to_port(interface_id)
                if not rc:
                    return ("MSG_OPS_PACKET_CAPTURE_FAILED_CONNECTION\n%s" %
                            interface_id)
            # Check params
            expression = self.get_arg_value_by_keyword("expression", cmd_list)
            capture_file = self.get_arg_value_by_keyword("file", cmd_list)
            file_count = self.get_arg_value_by_keyword("file-count", cmd_list)
            packet_count = self.get_arg_value_by_keyword("count", cmd_list)
            # Start capture
            result, err = start_packet_capture(
                interface_id, capture_to_file=True,
                expression=expression, capture_file=capture_file,
                packet_count=packet_count, direction=direction,
                dvfilter_name=dvfilter_name, dvfilter_stage=dvfilter_stage,
                snaplen=snaplen, file_count=file_count)
            # Destroy edge default session if any
            if g_node_type in [NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE, NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
                edge_session = PacketCaptureCommandRunner()
                if interface_id == edge_session.default_span_interface:
                    edge_session.destroy_default_edge_span_session()
            if result == CAPTURE_FAILED:
                if err:
                    return "MSG_OPS_PACKET_CAPTURE_ERROR_MESSAGE\n%s" % err
                return "MSG_OPS_PACKET_CAPTURE_ERROR"
            elif result == CAPTURE_INTERRUPT_CTRL_C:
                if err:
                    return "MSG_OPS_PACKET_INTERRUPT_CTRL_C_MESSAGE\n%s" % err
                return "MSG_OPS_PACKET_INTERRUPT_CTRL_C"
            return "OK\n\n%s" % simplejson.dumps("")
        finally:
            signal.signal(signal.SIGINT, sigint_handler)
            signal.signal(signal.SIGTERM, sigterm_handler)
            signal.signal(signal.SIGPIPE, sigpipe_handler)
            unlock_interface(lock_interface_id)
            if ((g_node_type == NODE_TYPE_HOST_SLESKVM) or
                (g_node_type == NODE_TYPE_HOST_CENTOSKVM) or
                (g_node_type == NODE_TYPE_HOST_RHELKVM) or
                (g_node_type == NODE_TYPE_HOST_UBUNTUKVM) or
                is_baremetal()):
                kvm_session.close_connection_to_port(interface_id)
                if interface_id == kvm_session.default_span_port:
                    kvm_session.destroy_default_kvm_span_session()



class SetPacketDisplay(CliCommandProvider):
    def __call__(self, cmd_list):
        try:
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            signal.signal(signal.SIGPIPE, signal.SIG_IGN)
            session = self.get_arg_value_by_keyword("session", cmd_list)
            dvfilter_name = self.get_arg_value_by_keyword("dvfilter", cmd_list)
            dvfilter_stage = self.get_arg_value_by_keyword("stage", cmd_list)
            direction = self.get_arg_value_by_keyword("direction", cmd_list)
            core = self.get_arg_value_by_keyword("core", cmd_list)
            if session:
                interface_id = "span-%s" % session
                lock_interface_id = interface_id
            else:
                interface_id = self.get_arg_value_by_keyword("interface", cmd_list)
                lock_interface_id = interface_id
                # For edge, we reserve session 0 to create span session
                if g_node_type in [
                        NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE, NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
                    if not lock_interface(lock_interface_id):
                        return ("MSG_OPS_PACKET_CAPTURE_ALREADY_STARTED_ERROR\n%s" %
                                lock_interface_id)
                    local_interfaces = get_interface_names()
                    if interface_id not in local_interfaces:
                        edge_session = PacketCaptureCommandRunner()
                        # there may be a stale session from previous pcap
                        # try at-least twice to create span session before returning error
                        # rc is true on success
                        for i in range(2):
                            rc, rv = edge_session.create_default_edge_span_session(
                                        interface_id, direction, core)
                            if rc:
                                break
                            edge_session.destroy_default_edge_span_session()
                        else:
                            return rv
                        interface_id = edge_session.default_span_interface
                    elif core:
                        return "MSG_SPAN_CORE_ERROR_LOCAL_INTERFACE"
                elif ((g_node_type == NODE_TYPE_HOST_SLESKVM) or
                      (g_node_type == NODE_TYPE_HOST_CENTOSKVM) or
                      (g_node_type == NODE_TYPE_HOST_RHELKVM) or
                      (g_node_type == NODE_TYPE_HOST_UBUNTUKVM) or
                      is_baremetal()):
                    local_interfaces = get_interface_names()
                    if interface_id not in local_interfaces:
                        kvm_session = PacketCaptureCommandRunner()
                        rv = (
                            kvm_session.create_default_kvm_span_session(
                                interface_id, direction))
                        if not rv.startswith("OK"):
                            return rv
                        interface_id = kvm_session.default_span_port
                elif g_node_type == NODE_TYPE_HOST_ESX:
                  if not interface_id:
                     if dvfilter_name:
                        # we treat dvfilter's name as interface id
                        interface_id = dvfilter_name
                     else:
                        # we treat "trace" as a special inteface id
                        interface_id = 'trace'
            # In KVM, open connection to the span_port
            if ((g_node_type == NODE_TYPE_HOST_SLESKVM) or
                (g_node_type == NODE_TYPE_HOST_CENTOSKVM) or
                (g_node_type == NODE_TYPE_HOST_RHELKVM) or
                (g_node_type == NODE_TYPE_HOST_UBUNTUKVM) or
                is_baremetal()):
                kvm_session = PacketCaptureCommandRunner()
                rc = kvm_session.open_connection_to_port(interface_id)
                if not rc:
                    kvm_session.close_connection_to_port(interface_id)
                    if interface_id == kvm_session.default_span_port:
                        kvm_session.destroy_default_kvm_span_session()
                    return ("MSG_OPS_PACKET_CAPTURE_FAILED_CONNECTION\n%s" %
                            interface_id)
            # Check params
            expression = self.get_arg_value_by_keyword("expression", cmd_list)
            packet_count = self.get_arg_value_by_keyword("count", cmd_list)
            # Start capture
            result, err = start_packet_capture(interface_id, capture_to_file=False,
                                               expression=expression,
                                               packet_count=packet_count,
                                               direction=direction,
                                               dvfilter_name=dvfilter_name,
                                               dvfilter_stage=dvfilter_stage)
            # Destroy edge or kvm default session if any
            if g_node_type in [NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE, NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
                edge_session = PacketCaptureCommandRunner()
                if interface_id == edge_session.default_span_interface:
                    edge_session.destroy_default_edge_span_session()
            elif ((g_node_type == NODE_TYPE_HOST_SLESKVM) or
                  (g_node_type == NODE_TYPE_HOST_CENTOSKVM) or
                  (g_node_type == NODE_TYPE_HOST_RHELKVM) or
                  (g_node_type == NODE_TYPE_HOST_UBUNTUKVM) or
                  is_baremetal()):
                kvm_session.close_connection_to_port(interface_id)
                if interface_id == kvm_session.default_span_port:
                    kvm_session.destroy_default_kvm_span_session()
            if result == CAPTURE_INVALID_INTERFACE:
                return "MSG_OPS_PACKET_CAPTURE_INVALID_INTERFACE"
            elif result == CAPTURE_FAILED:
                if err:
                    return "MSG_OPS_PACKET_CAPTURE_ERROR_MESSAGE\n%s" % err
                return "MSG_OPS_PACKET_CAPTURE_ERROR"
            elif result == CAPTURE_INTERRUPT_CTRL_C:
                if err:
                    return "MSG_OPS_PACKET_INTERRUPT_CTRL_C_MESSAGE\n%s" % err
                return "MSG_OPS_PACKET_INTERRUPT_CTRL_C"
            return "OK\n\n%s" % simplejson.dumps("")
        finally:
            unlock_interface(lock_interface_id)
            signal.signal(signal.SIGINT, sigint_handler)
            signal.signal(signal.SIGTERM, sigterm_handler)
            signal.signal(signal.SIGPIPE, sigpipe_handler)
