# ***************************************************************************
# Copyright 2014-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
import os
import yaml

from cli.descriptors.module_factory import ModuleFactory
from cli.descriptors.cli_argument_descriptor import CliArgumentDescriptor
from cli.descriptors.cli_command_descriptor import CliCommandDescriptor
from cli.descriptors.cli_cfg_serializer_descriptor \
    import CliCfgSerializerDescriptor
from cli.descriptors.cli_help_summary_descriptor import CliHelpSummaryDescriptor
from cli.descriptors.cli_message_descriptor import CliMessageDescriptor
from cli.descriptors.descriptor import Descriptor
from cli.descriptors.error import DescriptorException
from cli.descriptors.error import DescriptorTargetException
from cli.utils.cache_utils import get_nodes
from cli.utils.cache_utils import VerifyArgCache
from cli.utils.constants import CENTRAL_CLI_START_TOKEN
from cli.utils.constants import CENTRAL_CLI_END_TOKEN
from cli.utils.constants import CENTRAL_CLI_STR
from cli.utils.constants import MODE_ALL
from cli.utils.constants import NODE_TYPE_MANAGER
from cli.utils.constants import NON_QUIT_MODES
from cli.utils.constants import PERMISSION_CRUD
from cli.utils.constants import SHOW_CONFIG_DNS
from cli.utils.constants import SHOW_CONFIG_GLOBAL
from cli.utils.constants import SHOW_CONFIG_INTERFACE
from cli.utils.error_utils import set_messages
from cli.utils import node_utils
from cli.utils.error_code import ErrorCode as ec
from compat23.compat23_utils import compat23_string_types
import featureswitch


lg = nsx_logging.getLogger("cli.descriptors.descriptor_factory")


_DEFAULT_PERMISSIONS = PERMISSION_CRUD  # crud to maintain compatibility
# pre-Dropkick


_SUPPRESS_INVALID_PROVIDER_EXCEPTION = False


def suppress_invalid_providers_exception():
    """Specify that invalid providers should be ignored.

    For normal usage, the descriptor factory raises an exception if a CliCommand
    targeted for node type A is loaded on node type B. This causes problems when
    generating CLI documentation (becuase we have to load all CLICommand
    descriptors to know what documentation to generate). This method is used
    by the documentation generator to suppress and ignore descriptors as
    applicable.
    """
    global _SUPPRESS_INVALID_PROVIDER_EXCEPTION
    _SUPPRESS_INVALID_PROVIDER_EXCEPTION = True


class DescriptorFactory(object):
    """Factory class that loads and initializes all descriptors as defined
       in specified module files.
    """

    def __init__(self, module_files, settings_file=None,
                 prompt=None, user_permissions=_DEFAULT_PERMISSIONS,
                 enable_central_cli=False, **kwargs):
        self._prompt = prompt
        self._enabled_feature_switches = None
        self._descriptors = {}    # key=descriptor.id, val=Descriptor
        self._modules = {}        # key=module.id, val=Module
        self._arguments = {}      # key=argument name, val=CliArgumentDesciptor
        self._commands = {}       # key=context,
        # val={CliCommandDescriptor: path tuple}
        self._help_summary = {}   # key=help id, val=help summary
        self._error_msgs = {}     # key=error id, val=error message
        self._serializers = {}    # key=name, val=CliCfgSerializerDescriptor
        self._show_configs = {       # key=show config category,
            SHOW_CONFIG_DNS: [],     # val=list of descriptors that support
            SHOW_CONFIG_GLOBAL: [],  # showing config for the category
            SHOW_CONFIG_INTERFACE: []}
        self._user_permissions = user_permissions
        self._enable_central_cli = enable_central_cli
        self.central_cli_descriptors_loaded = False

        self._node_mapping = {
            'AutonomousEdgeNode': ('autonomous_edge.yml', ),
            'EdgeNode': ('edge.yml', ),
            'ESXI': ('host_esx.yml', ),
            'KeyNode': ('keymanager.yml', ),
            'KVM': ('host_kvm.yml', ),
            'ManagementClusterRoleConfig': ('controller.yml', 'manager.yml', 'policy.yml'),
            'PublicCloudGatewayNode': ('publiccloudgateway.yml', )
        }

        # Contains information about all cluster/fabric nodes in following form
        # This gets populated by CentralCLICacheThread from cli.utils.cache_utils
        # {
        #   '0c77a7c0-0b3e-11e7-b7a3-970a64d76956':
        #   {
        #       'uuid': 0C77A7C0-0B3E-11E7-B7A3-970A64D76956,
        #       'display_name': host1,
        #       'node_type': 'ESXI'
        #   }, ...
        # }
        self._node_cache = {}

        # Cache for the values of verify arguments type
        self.verify_arg_cache = VerifyArgCache()

        # Module factory provides interface to create new modules
        self._module_factory = ModuleFactory(
            self, pickle_dir=kwargs.get("pickle_dir"))

        if settings_file:
            if isinstance(settings_file, compat23_string_types):
                with open(settings_file) as fp:
                    Descriptor.settings = yaml.load(fp)
            # else it should be a file-like object
            else:
                Descriptor.settings = yaml.load(settings_file)

        if not isinstance(module_files, (list, tuple)):
            module_files = [module_files]

        self.role_file = module_files[0]
        module_files = set(module_files)

        for module_file in module_files:
            self._process_module_file(
                module_file, overwrite_version=kwargs.get("overwrite_version"))

        # TODO: Define CliHelpSummary descriptor in common/help_summary.yml
        self._help_summary["no"] = "Negate a command or set its defaults"
        self._help_summary[CENTRAL_CLI_START_TOKEN] = "Run Central CLI command"

        set_messages(self._error_msgs)

    def process_central_cli_descriptors(self, service=None):
        for node_type, module_files in list(self._node_mapping.items()):
            for module_file in module_files:
                file_path = os.path.join(os.path.dirname(
                    str(self.role_file)), module_file)
                self._process_module_file(file_path, node_type=node_type,
                                          service=service)
        self.central_cli_descriptors_loaded = True

    @property
    def node_cache_available(self):
        return len(self._node_cache) > 0

    @property
    def node_cache(self):
        # CentralCLICacheThread may not have initialized the node_cache (e.g. when
        # using -c option with nsxcli). Use get_nodes to make direct API call
        if (not self._node_cache and self._enable_central_cli):
            self._node_cache = get_nodes()
        return self._node_cache

    @node_cache.setter
    def node_cache(self, value):
        self._node_cache = value

    def _process_module_file(
            self, module_file, node_type=None, service=None, **kwargs):
        # In case of manager node, load all available pickles to support central
        # cli
        if NODE_TYPE_MANAGER in node_utils.get_node_type() or self._enable_central_cli:
            # PR: 2493493, loading pickles is managed by module factory
            # inorder to reuse the pickle data whenever required.
            module = self._module_factory.get_module(module_file,
                                                     load_all_pickles=True,
                                                     overwrite_version=kwargs.get("overwrite_version"))
        else:
            module = self._module_factory.get_module(module_file)

        if node_type:  # If loading for Central CLI, make module id unique
            module.meta.id = "%s%s" % (module.meta.id, node_type)
        if module.meta.id in self._modules:
            err = "Module '%s' imported multiple times" % module.meta.id
            raise DescriptorException(err)
        elif self.feature_disabled(module.meta):
            lg.debug(
                "Skipping module %s because its feature %s is disabled",
                module.meta.id, module.meta._nsx_feature)
            return
        self._modules[module.meta.id] = module

        for key, descriptor in list(module.descriptors.items()):
            if self.feature_disabled(descriptor):
                lg.debug(
                    "Skipping descriptor %s, id %s because its feature %s "
                    "is disabled", descriptor, descriptor.id,
                    descriptor._nsx_feature)
                continue
            else:
                # CliCommandDescriptor is being loaded as Central CLI process,
                # so it should contain which node_type are loading this. Also
                # ensure that descriptor.id ends with CENTRAL_CLI_STR.
                if node_type and isinstance(descriptor, CliCommandDescriptor):
                    try:
                        d = self._descriptors[key + CENTRAL_CLI_STR]
                    except KeyError:
                        d = descriptor
                        d.id = '%s%s' % (d.id, CENTRAL_CLI_STR)
                        d._path_tuple = ((CENTRAL_CLI_START_TOKEN,
                                          '<registered-node-uuid>') +
                                         d._path_tuple)
                        if d.doc_override:
                            d._doc_override = ('%s <registered-node-uuid> %s %s'
                                               % (CENTRAL_CLI_START_TOKEN,
                                                   CENTRAL_CLI_END_TOKEN,
                                                   d.doc_override))

                    # Set service object on command descriptor. Although service
                    # objects does this for non central cli descriptors. But
                    # while lazily loading them, we have to explicitly set this.
                    d.service = service

                    try:
                        d.node_types.append(node_type)
                    except AttributeError:
                        d.node_types = [node_type]

                    self._descriptors[d.id] = d
                    descriptor = d
                else:
                    if not isinstance(descriptor, CliCommandDescriptor):
                        self._descriptors[key] = descriptor
                    else:
                        # For non-Central CLI case, add to self._descriptors
                        # after self._process_command_descriptor is called
                        pass
            if isinstance(descriptor, CliMessageDescriptor):
                self._error_msgs[str(key)] = descriptor.message
            elif isinstance(descriptor, CliArgumentDescriptor):
                descriptor._factory = self
                self._arguments[descriptor.name] = descriptor
            elif isinstance(descriptor, CliHelpSummaryDescriptor):
                self._help_summary[descriptor.id] = descriptor.help_summary
            elif isinstance(descriptor, CliCfgSerializerDescriptor):
                descriptor._factory = self
                self._serializers[descriptor.name] = descriptor
            elif isinstance(descriptor, CliCommandDescriptor):
                self._process_command_descriptor(key, descriptor, node_type)
                # When loading CliCommandDescriptors for the non-Central CLI
                # use case, self._process_command_descriptor requires that the
                # current descriptor.id not exist in self._descriptors,
                # therefore, we add it to self._descriptors now.
                if not node_type:
                    self._descriptors[key] = descriptor

    def _process_command_descriptor(self, key, descriptor, node_type):
        descriptor.check_authorized(self._user_permissions)
        if descriptor.provider is None:
            if not _SUPPRESS_INVALID_PROVIDER_EXCEPTION:
                msg = ("Descriptor %s imported on unsupported node "
                       "type: %s" % (descriptor.id, node_utils.get_node_type()))
                lg.error(msg, error_code=ec.OPERATION_NOT_SUPPORTED)
                raise DescriptorTargetException(msg)
            return
        elif self.feature_disabled(descriptor):
            lg.debug(
                "Skipping descriptor %s, id %s because its feature %s "
                "is disabled", descriptor, descriptor.id,
                descriptor._nsx_feature)
            del self._descriptors[key]
            return
        descriptor._factory = self
        path = descriptor.raw_schema["path"]
        contexts = descriptor.raw_schema.get("context_available", [])
        contexts = (NON_QUIT_MODES
                    if not contexts or MODE_ALL in contexts else
                    contexts)
        add_to_show_configs = False
        for context in contexts:
            context_dict = self._commands.setdefault(context, {})
            if not node_type:
                # In the all-in-one appliance make sure the same descriptor
                # definition isn't added to self._commands and
                # self._show_configs more than once
                if descriptor.id not in self._descriptors:
                    context_dict[descriptor] = tuple(path)
                    add_to_show_configs = True
            elif (self._enable_central_cli and (
                  descriptor.id != CENTRAL_CLI_STR)):
                k = self._descriptors[descriptor.id]
                context_dict[k] = k._path_tuple
        if add_to_show_configs:
            category = descriptor.show_configuration_category
            if category == SHOW_CONFIG_DNS:
                self._show_configs[SHOW_CONFIG_DNS].append(descriptor)
            elif category == SHOW_CONFIG_GLOBAL:
                self._show_configs[SHOW_CONFIG_GLOBAL].append(descriptor)
            elif category == SHOW_CONFIG_INTERFACE:
                self._show_configs[SHOW_CONFIG_INTERFACE].append(descriptor)

    @property
    def descriptors(self):
        return self._descriptors

    @property
    def modules(self):
        return self._modules

    @property
    def arguments(self):
        return self._arguments

    @property
    def commands(self):
        return self._commands

    @property
    def help_summary(self):
        return self._help_summary

    @property
    def show_config_commands(self):
        return self._show_configs

    def feature_disabled(self, descriptor):
        # Populate enabled feature switches list if needed
        if self._enabled_feature_switches is None:
            self._enabled_feature_switches = []
            featureswitch.feature_switch_init()
            for switch_name, enabled in list(
                    featureswitch.get_switches().items()):
                if enabled:
                    self._enabled_feature_switches.append(switch_name)
        if (not hasattr(descriptor, "_nsx_feature") or
                descriptor._nsx_feature is None):
            return False
        elif isinstance(descriptor._nsx_feature, str):
            return descriptor._nsx_feature not in self._enabled_feature_switches
        result = True
        for feature in descriptor._nsx_feature:
            if feature in self._enabled_feature_switches:
                result = False
                break
        return result
