# ***************************************************************************
# Copyright 2014-2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************

__all__ = ["ModuleFactory"]

import os
import sys
from cli.descriptors.module_descriptor import Module
from cli.utils.constants import PICKLE_DUMPS_DIR as _PICKLE_DUMPS_DIR
from cli.utils.constants import PICKLE_EXTENSION as _PICKLE_EXTENSION


class ModuleFactory(object):
    """Factory class that provides interface to create new module,
    where module is built with available pickle file."""

    # PR: 2493493, pickle files should be loaded & stored once,
    # redundant copies consumes high memory in cli.

    _pickles = dict()

    def __init__(self, descriptor,
                 pickle_dir=None, pickle_ext=_PICKLE_EXTENSION):
        self._pickle_dir = _PICKLE_DUMPS_DIR if pickle_dir is None else pickle_dir
        self._pickle_ext = pickle_ext
        self._descriptor = descriptor

    def get_module(self, module_file, load_all_pickles=False, **kwargs):
        """
        Creates new module as per module file.
        Pickles are reused while creating new module.

        Args:
            module_file: (str) path of module.
            load_all_pickles: (bool) flag when enabled loads all available pickle files with module.
                                required in case of manager nodes, which supports Central Cli.
        """
        module = Module(
            module_file,
            self._descriptor,
            overwrite_version=kwargs.get("overwrite_version"))

        if load_all_pickles:
            module._add_pickle_dump(self._get_all_pickle_dumps())
        return module

    def _get_all_pickle_dumps(self):
        """Retrives all pickle dumps data, available in pickle directory"""
        data = dict()
        if os.path.isdir(self._pickle_dir):
            file_names = os.listdir(self._pickle_dir)
            for file_name in file_names:
                if os.path.splitext(file_name)[-1] != self._pickle_ext:
                    continue
                pickle_path = os.path.join(self._pickle_dir, file_name)
                pickle_key = os.path.split(pickle_path)[-1]

                data[pickle_key] = self._get_pickle_dump(pickle_path)
        return data

    def _get_pickle_dump(self, pickle_path):
        """Retrives pickle dump of a specified pickle path.
        Loads pickle file only once & is reused for next references.
        """
        pickle_key = os.path.split(pickle_path)[-1]
        if pickle_key not in self._pickles:
            data = Module._load_pickle_file(pickle_path)
            self._pickles[pickle_key] = data
        return self._pickles[pickle_key]
