# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential

import ipaddress
import json
import os.path
import re
import traceback

from sha.contrib.metric.edge_base import EdgeBase
from sha.contrib.metric.utils import timeout_cache

from subprocess import CalledProcessError

# tunnel status definition
# 'unknown' means other status than known status from command
IPSEC_TUNNEL_STATUS_INVALID = 'invalid'
IPSEC_TUNNEL_STATUS_UP = 'up'
IPSEC_TUNNEL_STATUS_DOWN = 'down'
IPSEC_TUNNEL_STATUS_DELETE= 'delete'
IPSEC_TUNNEL_STATUS_UNKNOWN = 'unknown'

IPSEC_STATUS_INVALID_STR = 'IPSEC_STATUS_INVALID'
IPSEC_STATUS_UP_STR = 'IPSEC_STATUS_UP'
IPSEC_STATUS_DOWN_STR = 'IPSEC_STATUS_DOWN'
IPSEC_STATUS_DELETE_STR = 'IPSEC_STATUS_DELETE'

tunnel_status_map = {
    IPSEC_STATUS_INVALID_STR : IPSEC_TUNNEL_STATUS_INVALID,
    IPSEC_STATUS_UP_STR : IPSEC_TUNNEL_STATUS_UP,
    IPSEC_STATUS_DOWN_STR : IPSEC_TUNNEL_STATUS_DOWN,
    IPSEC_STATUS_DELETE_STR : IPSEC_TUNNEL_STATUS_DELETE
}

# Refer to nsx/edge/iked/src/iked_event.h and iked_event.c
session_status_map = {
    "IKE_STATUS_INVALID" : 0,
    "IKE_STATUS_UP": 1,
    "IKE_STATUS_DOWN" : 2,
    "IKE_STATUS_NEGO" : 3,
    "IKE_STATUS_DELETE" : 4
};
# Default zero tunnel stats template
# Considering there maybe no stats of any tunnel can be get on an Edge node,
# we need a pre-defined stats template.
# Note: the stats template need be updated once any adjustment of tunnel stats.
default_tunnel_stats_template = {
   "tx_bytes":0,
   "rekey_request_failure":0,
   "tx_tcp_mss_ignored":0,
   "rx_drops":0,
   "rx_drop_auth_failure":0,
   "rx_drop_crypto_failure":0,
   "rx_drop_pol_nomatch":0,
   "rx_drop_blocked":0,
   "tx_tcp_mss_adjusted":0,
   "rx_tcp_mss_adjusted":0,
   "tx_drop_blocked":0,
   "tx_drop_malformed":0,
   "tx_packets":0,
   "rx_drop_malformed":0,
   "tx_drop_proto_not_supported":0,
   "rx_drop_proto_not_supported":0,
   "tx_drops":0,
   "rx_packets":0,
   "tx_drop_state_mismatch":0,
   "tx_drop_frag_needed":0,
   "rx_drop_state_mismatch":0,
   "tx_drop_seq_rollover":0,
   "rx_drop_replay":0,
   "rx_bytes":0,
   "tx_drop_crypto_failure":0,
   "rx_tcp_mss_ignored":0,
   "rx_drop_inner_malformed":0
}

def int_to_ipv4_str(ip_number):
    return ipaddress.IPv4Address(ip_number).exploded

class EdgeIpsecVpnStats(EdgeBase):
    def __init__(self):
        super().__init__()
        self.tunnel_stats_template = None

    def get_tunnel_stats(self, tunnel_id, not_active):
        # Get one tunnel stats
        cmd_get_tunnel = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'ipsec_tunnel/show', 'verbose', tunnel_id]
        try:
            output = self.run_command_with_time_check(cmd_get_tunnel, expected_failure = not_active)
        except CalledProcessError as e:
            self._logger.info("Failed to get tunnel stats of {}, report all zero stats".format(tunnel_id))
            return default_tunnel_stats_template if not self.tunnel_stats_template else self.tunnel_stats_template

        stats  = json.loads(output)['stats']
        self._logger.debug("get edge vpn tunnel {} :{}".format(tunnel_id, stats))

        # stats key conveting: rx_pkts/tx_pkts to rx_packets/tx_packets
        stats['rx_packets'] = stats.pop('rx_pkts', 0)
        stats['tx_packets'] = stats.pop('tx_pkts', 0)

        if not self.tunnel_stats_template:
            self.tunnel_stats_template = stats.copy()
            for key in self.tunnel_stats_template.keys():
                self.tunnel_stats_template[key] = 0
            self._logger.info("set tunnel_stats_template {}".format(self.tunnel_stats_template))

        return stats

    def get_vti_iface_stats(self, vti_id):
        # Get VTI interface stats
        # Cost of nsxcli is much higher than edge backend command
        #     """nsxcli -c 'get logical-router interface {} stats | json'"""

        cmd_get_iface = ['edge-appctl', '-t', '/var/run/vmware/edge/dpd.ctl', 'lrouter_port/show', vti_id]
        output = self.run_command_with_time_check(cmd_get_iface)

        iface_stats = json.loads(output)['stats']
        self._logger.debug("get iface vti {}: {}".format(vti_id, iface_stats))
        # stats key conveting: rx_pkts/tx_pkts to rx_packets/tx_packets
        iface_stats['rx_packets'] = iface_stats.pop('rx_pkts', 0)
        iface_stats['tx_packets'] = iface_stats.pop('tx_pkts', 0)
        return iface_stats

    def aggregate_vpn_stats(self, session_stats, tunnel_stats):
        for key, value in tunnel_stats['tunnel_status'].items():
            session_stats['tunnel_status'][key] += value
        for key, value in tunnel_stats.get('tunnel_stats', {}).items():
            session_stats['tunnel_stats'][key] = session_stats['tunnel_stats'].get(key, 0) + value
        self._logger.debug("Aggregating session stats: {}".format(session_stats))
        return session_stats

    def get_vpn_stats(self, lr_uuid):
        # When ipsecvpn service is disabled, there is no ike.ctl. Check it firstly
        if not os.path.exists("/var/run/vmware/edge/ike.ctl"):
            self._logger.info("No ike.ctl file, the ike service is disabled.")
            return None

        # Get vpn sessions
        #
        # Cost of nsxcli is much higher than edge backend command
        #     """nsxcli -c 'get ipsecvpn session logical-router {} | json'"""
        cmd_get_sessions = ['edge-appctl', '-t', '/var/run/vmware/edge/ike.ctl', 'vpn-session/get', 'sr-id', lr_uuid]
        output = self.run_command_with_time_check(cmd_get_sessions)

        vpn_sessions = json.loads(output)
        self._logger.debug("get edge vpn sessions {}".format(vpn_sessions))


        # Nsxcli can get all tunnels' stats in one command
        #     """nsxcli -c 'get ipsecvpn tunnel stats | json'"""
        # But cost of nsxcli is much higher than edge backend command
        # Will get tunnel one-by-one via backend command later

        stats_list = []
        for session in vpn_sessions:
            self._logger.debug("session {}: enabled({}), HA({})".format(session['id'], session['Enabled'], session['HA_Status']))

            # Currently, session only support IPv4 address
            session_name = int_to_ipv4_str(session['Local_Endpoint_Profile']['Local_Address']['ipv4']).replace('.', '-') + '_' + \
                           session['Peer_Endpoint_Profile']['Peer_Address'].replace('.', '-')
            if session['Type'] == 'POLICY_BASED_SESSION':
                type = 'pbvpn'
            elif session['Type'] == 'ROUTE_BASED_SESSION':
                type = 'rbvpn'
            else:
                self._logger.warning("Unknown session type of {}: {}".format(session['id'], session['Type']))
                continue

            session_stats = {
                'tunnel_status' : {
                    IPSEC_TUNNEL_STATUS_INVALID : 0,
                    IPSEC_TUNNEL_STATUS_UP : 0,
                    IPSEC_TUNNEL_STATUS_DOWN : 0,
                    IPSEC_TUNNEL_STATUS_DELETE : 0,
                    IPSEC_TUNNEL_STATUS_UNKNOWN : 0
                },
                'tunnel_stats' :{}
            }
            iface_stats = {}
            for policy in session['Policy']:
                try:
                    stats = self.get_tunnel_stats(policy['id'], session['HA_Status'] != 'Active')
                except Exception as e:
                    trace = traceback.format_exc()
                    self._logger.warning("Failed to get tunnel stats of {}: {} {}".format(policy['id'], e, trace))
                    stats = {}

                # Save tunnel_stats
                tunnel_stats = {
                    'tunnel_status': {
                        tunnel_status_map.get(policy['Tunnel_Status'], IPSEC_TUNNEL_STATUS_UNKNOWN): 1
                    }
                }
                if stats:
                    tunnel_stats['tunnel_stats'] = stats

                session_stats = self.aggregate_vpn_stats(session_stats, tunnel_stats)

                if type == 'rbvpn':
                    try:
                        iface_stats = self.get_vti_iface_stats(policy['id'])
                    except Exception as e:
                        trace = traceback.format_exc()
                        self._logger.warning("Failed to get iface stats on VTI {}: {}, {}".format(policy['id'], e, trace))

                    # RBVPN, only one tunnel in each session
                    break

            session_stats['ike_sa_status'] = session_status_map.get(session["Session_Status"], -1)
            self._logger.debug("Aggregated session stats: {}".format(session_stats))

            vpn_stats = {}
            if session_stats:
                vpn_stats['vpn.' + type] = {session_name: session_stats}
            if iface_stats:
                vpn_stats['iface.vti'] = {session_name: iface_stats}
            if vpn_stats:
                stats_list.append(vpn_stats)

        self._logger.debug("session stats: {}".format(stats_list))
        return None if not stats_list else stats_list


    @timeout_cache(10)
    def usage(self, *args, **kwargs):
        lr_dict = {}
        try:
            lr_dict = self.get_lr_list(['SERVICE_ROUTER_TIER0', 'SERVICE_ROUTER_TIER1'])
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get edge lrouter list: {}, {}".format(e, trace))
        finally:
            if not lr_dict:
                return None

        lr_stats_list = []
        for key,lr in lr_dict.items():
            try:
                stats = self.get_vpn_stats(key)
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to get edge vpn stats on {}: {}, {}".format(lr, e, trace))
                stats = None

            if stats is None:
                continue

            lr_stats_list.append({'name': lr['type'] + '.' +lr['name'], 'vpn_stats': stats})

        return None if not lr_stats_list else (self.wavefront_source, lr_stats_list)
