# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential
import json
import os
import platform
import operator
import subprocess
import traceback
import sha.config
from sha.contrib.metric.abc_base import Metric
from sha.contrib.metric.utils import run_command, timeout_cache


class HyperbusConnectionList(Metric):
    FULL_VIF_STATE = 'LSP_ATTACHMENT/VIF_CONNECT/VIF_HYPERBUS_ENABLE'
    ALL_ZERO_IP = '0.0.0.0'
    STABLE = 'stable'

    def __init__(self):
        self._logger = sha.config.get_logger(__name__)

    @timeout_cache(2)
    def usage(self, *args, **kwargs):
        def get_vif_state_map_and_unhealthy_connection(vif_lips):
            """
            1. convert list to vif_state_map - { VIFID : State }
            2. get UNHEALTHY connections (this UNHEALTHY case only occurs on Esx host)
               - State:LSP_ATTACHMENT/VIF_CONNECT/VIF_HYPERBUS_ENABLE && IP == 0.0.0.0
            """
            vif_state_map = dict()
            unhealthy_connections = []

            for vif_lip in vif_lips:
                vif_id = vif_lip.get('VIFID', None)
                if vif_id is not None:
                    vif_id = vif_id.strip()

                if not vif_id:
                    self._logger.warn('VIFID is invalid, vif_lip:{}'.format(vif_lip))
                    continue

                state = vif_lip.get('State', None)
                ip = vif_lip.get('IP', None)

                if state is not None:
                    state = state.strip()
                vif_state_map[vif_id] = state

                if ip is not None and state is not None and platform.system() == 'VMkernel':
                    ip = ip.strip()
                    if ip == self.ALL_ZERO_IP and state == self.FULL_VIF_STATE:
                        # PR 2527833, this UNHEALTHY case only occurs on Esx host.
                        connection = {'VIFID': vif_id,
                                      'Status': 'UNHEALTHY',
                                      'Detail': 'Vmknic config is incorrect',
                                      self.STABLE: False}
                        unhealthy_connections.append(connection)

            self._logger.debug(
                'got vif_state_map: {}, unhealthy_connections:{}'.format(vif_state_map, unhealthy_connections))
            return vif_state_map, unhealthy_connections

        def update_connection_status(connections, vif_lips):
            """
            For each connection,
            If connection status is HEALTHY, then keep it as HEALTHY.
            For those non-HEALTHY status connections, if its vif-lip-state is 'LSP_ATTACHMENT/VIF_CONNECT/VIF_HYPERBUS_ENABLE',
            then set its connection status as UNHEALTHY, otherwise HEALTHY.

            Add those UNHEALTHY connections from vif-lip/table.
            """
            vif_state_map, unhealthy_connections = get_vif_state_map_and_unhealthy_connection(vif_lips)
            for connection in connections:
                vif_id = connection.get('VIFID', None)
                if vif_id is not None:
                    vif_id = vif_id.strip()

                if not vif_id:
                    self._logger.warn('VIFID is invalid, connection:{}'.format(connection))
                    continue
                connection['VIFID'] = vif_id

                status = connection.get('Status', None)
                if status is not None:
                    status = status.strip()
                connection['Status'] = status

                if status == 'HEALTHY':
                    connection[self.STABLE] = True
                    continue
                if status == 'INITIALIZING':
                    connection[self.STABLE] = True
                    continue
                if vif_state_map.get(vif_id, None) != self.FULL_VIF_STATE:
                    connection[self.STABLE] = True
                    continue
                connection[self.STABLE] = False

            connections.extend(unhealthy_connections)
            self._logger.debug('update_connection_status, connections:{}'.format(connections))

        try:
            system_name = platform.system()
            if system_name == 'VMkernel':
                """
                [{"HostSwitchID":"e0 e9 20 07 32 a3 44 c3-b6 70 54 55 78 57 39 7e","Status":"HEALTHY","VIFID":"314ce210-d0b4-4640-bcfa-85e47735a21c","Connection":"169.254.2.14:2345"},
                {"HostSwitchID":"e0 e9 20 07 32 a3 44 c3-b6 70 54 55 78 57 39 7e","Status":"COMMUNICATION_ERROR","VIFID":"32f02662-5737-4a73-a3bb-1fd763ab4d99","Connection":"169.254.2.18:2345"},
                {"HostSwitchID":"ff 3e c5 32 38 e7 4e 7a-aa 96 7d 12 48 fc 72 5e","Status":"INITIALIZING","VIFID":"af55c890-bdb2-43a4-80c2-d64e5ee935a9","Connection":"169.254.1.10:2345"}]
                """
                vif_connections_cmd = ["/opt/vmware/nsx-cli/bin/nsx-appctl", "-t", "/var/run/vmware/hyperbus/hyperbus-cli", "vif/connections"]
                """
                [{"State":"LSP_ATTACHMENT/VIF_CONNECT/VIF_HYPERBUS_ENABLE","IP":"169.254.2.14","VIFID":"314ce210-d0b4-4640-bcfa-85e47735a21c"},
                {"State":"LSP_ATTACHMENT/VIF_CONNECT/VIF_HYPERBUS_ENABLE","IP":"169.254.2.18","VIFID":"32f02662-5737-4a73-a3bb-1fd763ab4d99"},
                {"State":"LSP_ATTACHMENT/VIF_HYPERBUS_ENABLE","IP":"169.254.1.10","VIFID":"af55c890-bdb2-43a4-80c2-d64e5ee935a9"}]
                """
                vif_lip_cmd = ["/opt/vmware/nsx-cli/bin/nsx-appctl", "-t", "/var/run/vmware/hyperbus/hyperbus-cli", "vif-lip/table"]
            elif system_name == "Linux":
                """
                [{"Status":"MISS_VERSION_HANDSHAKE","VIFID":"904796a5-0e1a-4494-851a-04d5bb36da9b","Connection":"169.254.1.10:2345"}]
                """
                vif_connections_cmd = ["/usr/bin/nsx-appctl", "-t", "/var/run/vmware/hyperbus/hyperbus-cli", "vif/connections"]
                """
                [{"VIFID":"904796a5-0e1a-4494-851a-04d5bb36da9b","IP":"169.254.1.10","State":"LSP_ATTACHMENT/VIF_CONNECT/VIF_HYPERBUS_ENABLE"}]
                """
                vif_lip_cmd = ["/usr/bin/nsx-appctl", "-t", "/var/run/vmware/hyperbus/hyperbus-cli", "vif-lip/table"]
            else:
                return []

            try:
                connections = run_command(cmd=vif_connections_cmd, shell=False)
            except subprocess.CalledProcessError as e:
                trace = traceback.format_exc()
                self._logger.error(
                    'Failed in getting connections by cmd:{}, exception:{}, trace:{}'.format(vif_connections_cmd, e, trace))
                return None

            try:
                vif_lips = run_command(cmd=vif_lip_cmd, shell=False)
            except subprocess.CalledProcessError as e:
                trace = traceback.format_exc()
                self._logger.error(
                    'Failed in getting vif lip table by cmd:{}, exception:{}, trace:{}'.format(vif_lip_cmd, e, trace))
                return None

            connections = json.loads(connections)
            self._logger.debug('got connections: {}'.format(connections))
            vif_lips = json.loads(vif_lips)
            self._logger.debug('got vif_lips: {}'.format(vif_lips))

            update_connection_status(connections, vif_lips)

            hyperbus_connection_list = []

            for connection in connections:
                hyperbus_connection = {'id': connection['VIFID'], 'status': connection['Status'],
                                       self.STABLE: connection[self.STABLE]}
                if 'Detail' in connection:
                    hyperbus_connection['detail'] = connection['Detail']
                hyperbus_connection_list.append(hyperbus_connection)

            self._logger.debug('got list {}'.format(hyperbus_connection_list))
            # As we check the report_msg (calculated basing on this list) diff by '==', here sort the list by id,
            # so that if there is no data change in list, current report_msg is always equals with pre-one.
            # i.e. order change among list items won't be treated as data change
            hyperbus_connection_list.sort(key=operator.itemgetter('id'))
            return hyperbus_connection_list
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.error(
                'Unexpected error in get hyperbus connections, exception:{}, trace:{}'.format(e, trace))
            # if encounter unexpected error, here should return empty list as this is
            # a SHA issue not a hyperbus issue, so that for this case:
            # 1. hyperbus monitor should report healthy to Manager; Trigger Alarm=False
            # 2. node agent monitor should report empty msg to Manager; Trigger Alarm=False
            return []



