# Copyright (C) 2020-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
import time
import uuid
from sys import getsizeof, stderr
from itertools import chain
from collections import deque
from reprlib import repr


def convert_to_uuid_str(uuid_msg):
    """
    convert UuidMsg to uuid string
    :param uuid_msg: id {   left: 6058522876635726111   right: 9329288385188951827 }
    :return: uuid string: '54143274-f573-411f-8178-4a5b8e5d7713'
    """
    uuid_int = uuid_msg.left << 64 | uuid_msg.right
    return str(uuid.UUID(int=uuid_int))


def total_size(o, handlers={}, verbose=False):
    """ Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    """
    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {tuple: iter,
                    list: iter,
                    deque: iter,
                    dict: dict_handler,
                    set: iter,
                    frozenset: iter,
                   }
    all_handlers.update(handlers)     # user handlers take precedence
    seen = set()                      # track which object id's have already been seen
    default_size = getsizeof(0)       # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:       # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)


def sleep_and_check_running(t, seconds):
    """
    use this to make current thread sleep. Current thread will stop sleep when either of below happens
    1. current thread stop running.
    2. the sleep time reached.
    :param t: thread to sleep. It must have a property 'is_running'
    :param seconds: time to sleep
    :return:
    """
    sleep_seconds = seconds
    # check the running flag to exit in time
    while t.is_running and sleep_seconds > 0:
        time.sleep(0.1)
        sleep_seconds -= 0.1
