# The interface module, which contains the the protobuf messages used to
# generate the interface commands.
from vmware.nsx.cli import metrics_cli_pb2

# Additional modules used in custom functions.
import interface
import sys
from collections import defaultdict

# An interface for getting nsx-proxy metrics data. Serves as an example
# of an interface containing both custom and generated methods.
class Metrics(interface.Interface):

    def __init__(self):
        # Required initialization:
        interface.Interface.__init__(self, metrics_cli_pb2, metrics_cli_pb2.MetricsCliService_Stub)

        # To use the default help message for custom commands, you must call
        # make_default_help in an interface module's __init__, with an optional
        # 'info' string to further describe the custom method.

        # To have a custom help function for a custom command, define a
        # function 'help_CommandName' which corresponds to your custom
        # get_CommandName.
        interface.Interface.make_default_help(self, "GetMetrics", "MetricTable")

        tree_info = "Returns a visualization of the metrics tree structure starting from" + \
        "below the specified prefix. The number of '-' corresponds to the depth" + \
        "of the entry in the tree."
        interface.Interface.make_default_help(self, "GetMetrics", "MetricTree", info=tree_info)

    # This command does exactly what GetMetrics does, but with some additional
    # formatting to print a table. In order to generate a CLI command for it,
    # the function must start with 'get_'. Helper methods meant for internal
    # interface use should NOT start with 'get_'.
    def get_MetricTable(self, args):

        # This method should be called at the beginning of every custom function.
        interface.Interface.err_check(self)

        # Writing a custom method requires preparing a request message and
        # calling a protobuf rpc method that takes that message. Here we're using:

        #   service MetricsCliService {
        #       rpc GetMetrics(MetricRequestMsg) returns (MetricResponseMsg);
        #   }
        # See: nsx/proto_oh/vmware/nsx/lcp/metrics_cli.proto

        # This method starts with 'Get', so a 'Metrics' function is also
        # generated automatically, but protobuf messages used in custom
        # functions do not need to start with 'Get' (and shouldn't, if you do
        # not want an additional generated function).

        # Create a MetricRequestMsg and populate its 'prefix' field with the
        # input argument.
        req = metrics_cli_pb2.MetricRequestMsg()
        if args:
            if len(args) > 1:
                return "Error: too many arguments!"
            if args[0] != '/':
                args[0] = '/' + args[0]
            req.prefix = args[0]
        else:
            req.prefix = '/nsx'

        # We use the multi-stub syntax (e.g. ServiceName_MethodName(request))
        # to retrieve our response. Interface methods all run through the
        # Client Engine, which always expects multi-stub syntax.
        # See: nsx/common/rpc/src/python/vmware/nsx/rpc/client/client.py
        rsp = self.client.MetricsCliService_GetMetrics(req)

        # Response formatting:
        table = []
        columns = ["postfix"]
        maxlen = defaultdict(int)
        maxlen["postfix"] = len("postfix")
        for m in rsp.metric:
            entry = defaultdict(str)
            postfix = m.key.split(req.prefix)[-1].strip()
            entry["postfix"] = str(postfix)
            maxlen["postfix"] = max(maxlen["postfix"], len(postfix))

            vals = m.value.encode('ascii', 'replace').split(", ")
            for v in vals:
                s = v.split(":")
                s = [i.strip() for i in s]

                # /nsx/net/acceptor/state containts 2 entries that contain only
                # ['Listening'], so we handle these two entries specially
                # whenever we query the part of the metric tree that these
                # entries are in
                if len(s) <= 1:
                    maxlen[s[0]] = len(s[0])
                    entry["state"] = s[0]
                    table.append(entry)
                    if "state" not in columns:
                        columns.append("state")
                    continue

                entry[s[0]] = s[1]
                maxlen[s[0]] = max(maxlen[s[0]], len(s[1]), len(s[0]))
                table.append(entry)
                if s[0] not in columns:
                    columns.append(s[0])

        headers = ' | '.join([c.ljust(maxlen[c]) for c in columns])
        linebreak =  '-' * len(headers)
        data = '\n'.join([' | '.join([t[c].ljust(width)
            for (c, width) in zip(columns, [maxlen[c]
                for c in columns])]) for t in table])
        return headers + '\n' + linebreak + '\n' + data + '\n'

    # Another custom function
    def get_MetricTree(self, args):
        interface.Interface.err_check(self)

        req = metrics_cli_pb2.MetricRequestMsg()
        if args:
            if len(args) > 1:
                return "Error: too many arguments!"
            if args[0] != '/':
                args[0] = '/' + args[0]
            req.prefix = args[0]
        else:
            req.prefix = '/'

        rsp = self.client.MetricsCliService_GetMetrics(req)

        def process_single(path, tree):
            if len(path) == 0:
                return tree

            if len(path) == 1:
                # don't include id=, priority=, etc.
                if "=" in path[0]:
                    return tree
                else:
                    tree[path[0]] = {}
                    return tree

            if path[0] not in tree:
                tree[path[0]] = {}
            process_single(path[1:], tree[path[0]])

        def recursive_print(src, dpth = 0, key = ''):
            """ Recursively prints nested elements."""
            tabs = lambda n: ' ' * n * 4 # or 2 or 8 or...
            dash = lambda s, n: '%s%s' % ('-'*(n+1)+' ', s)

            for key, value in src.iteritems():
                print(tabs(dpth) + dash(key, dpth))
                recursive_print(value, dpth + 1, key)
            return ''

        paths = []
        for m in rsp.metric:
            postfix = m.key.split(req.prefix, 1)[-1].strip()
            path = postfix.split('/')
            paths.append(path[1:])

        tree = {}
        for path in paths:
            process_single(path, tree)

        return recursive_print(tree)
